/*******************************************************************************
 * Copyright (c) 2003, 2006 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.core.resources.restorer;

import java.io.DataInputStream;
import java.io.IOException;
import org.eclipse.core.internal.localstore.SafeChunkyInputStream;
import org.eclipse.core.resources.*;
import org.eclipse.core.runtime.*;
import org.eclipse.core.tools.restorer.IRestorer;
import org.eclipse.core.tools.restorer.RestorerPlugin;

public class ResourcesRestorer implements IRestorer {

	/** The project location file name. */
	private final static String LOCATION_FILE_NAME = ".location"; //$NON-NLS-1$

	/** The projects' metadata base path (relative to this plug-ins state location) */
	private final static String PROJECTS_METADATA_PATH = ".projects"; //$NON-NLS-1$

	/** The workspace */
	private static IWorkspace workspace = ResourcesPlugin.getWorkspace();

	/**
	 * Helper method that creates a IStatus object given a message and a Throwable
	 * object.
	 * 
	 * @param message a human-readable message 
	 * @param exception a low-level exception, or <code>null</code> if not
	 *    applicable 
	 * @return the created status object
	 */
	protected static IStatus createStatus(String message, Throwable exception) {
		return new Status(IStatus.ERROR, RestorerPlugin.PI_RESTORER, IStatus.ERROR, message, exception);
	}

	/**
	 * @see IRestorer#restoreFrom(IPath)
	 */
	public IStatus restoreFrom(IPath oldMetadataArea) {

		final MultiStatus result = new MultiStatus(RestorerPlugin.PI_RESTORER, IStatus.OK, "Problems restoring projects.", null); //$NON-NLS-1$

		java.io.File oldProjectsRootDir = new java.io.File(oldMetadataArea.toFile(), PROJECTS_METADATA_PATH);

		// restores each project that has metadata information available    		
		final java.io.File[] projectMetaDirs = oldProjectsRootDir.listFiles();

		// if there is no projects' metadata to restore, ignore     
		if (projectMetaDirs == null || projectMetaDirs.length == 0)
			return result;

		// wrap all the project creations in a workspace runnable to reduce notifications
		IWorkspaceRunnable job = new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) {
				for (int i = 0; i < projectMetaDirs.length; i++) {
					String projectName = projectMetaDirs[i].getName();
					IPath projectLocation = null;
					try {
						projectLocation = getProjectLocation(projectMetaDirs[i]);
					} catch (IOException e) {
						result.add(createStatus("Error reading location file for project: " + projectName, e)); //$NON-NLS-1$
						continue;
					}
					// create the project
					createProject(projectName, projectLocation);
				}
			}
		};
		
		try {
			workspace.run(job, null);
		} catch (CoreException e) {
			// core exceptions have already been caught and handled in the runnable
		}

		// after re-creating the projects we need to save the workspace
		try {
			workspace.save(true, null);
		} catch (CoreException e) {
			result.add(createStatus("Error saving workspace.", e)); //$NON-NLS-1$
		}

		return result;
	}

	/**
	 * Returns the project non-default location (null if default). A non-default 
	 * project location is obtained by reading the .location file in the project's 
	 * metadata directory.  
	 * 
	 * @param projectMetaDir a File pointing to the project's metadata directory
	 * @return the projects non-default location, or null if a default 
	 * location is used
	 * @throws IOException if an exception occurs while reading the location file
	 */
	protected IPath getProjectLocation(java.io.File projectMetaDir) throws IOException {
		java.io.File locationFile = new java.io.File(projectMetaDir, LOCATION_FILE_NAME);
		if (!locationFile.isFile())
			return null;
		DataInputStream is = null;
		try {
			// this file is written encoded with SafeChunkyOutputStream
			is = new DataInputStream(new SafeChunkyInputStream(locationFile));
			return new Path(is.readUTF());
		} finally {
			if (is != null)
				is.close();
		}
	}

	/**
	 * Creates a new open project with a project description built from 
	 * the arguments given. Does nothing if project already exists.
	 * 
	 * @param projectName the name of the project being recovered
	 * @param projectLocation the non-default project's location (null if default location)
	 * @see IWorkspace#newProjectDescription(String)
	 * @see IProjectDescription#setLocation(IPath)
	 * @see org.eclipse.core.resources.IProject#create(IProjectDescription, IProgressMonitor)
	 */
	protected IStatus createProject(String projectName, IPath projectLocation) {
		MultiStatus result = new MultiStatus(RestorerPlugin.PI_RESTORER, IStatus.OK, "Problems creating projects from: " + projectLocation, null); //$NON-NLS-1$
		IProject project = workspace.getRoot().getProject(projectName);
		// if project already exists, there is nothing to do
		if (project.exists())
			return result;
		// otherwise create the project in the workspace
		IProjectDescription description = workspace.newProjectDescription(projectName);
		description.setLocation(projectLocation);
		try {
			project.create(description, null);
			project.open(null);
		} catch (CoreException e) {
			result.add(createStatus("Exception creating and opening project at: " + projectLocation, e)); //$NON-NLS-1$
		}
		return result;
	}

}
