/*******************************************************************************
 * Copyright (c) 2003, 2006 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.core.tools.restorer;

import java.util.*;
import org.eclipse.core.runtime.*;
import org.eclipse.osgi.util.ManifestElement;
import org.osgi.framework.*;

/**
 * Attempts to restore the user's workspace.
 */
public class Restorer implements IPlatformRunnable {

	// create the comparator for sorting the plug-in descriptors.
	// it is based on the pre-requisite order.
	Comparator comparator = new Comparator() {
		public int compare(Object o1, Object o2) {
			Bundle base = (Bundle) o1;
			Bundle other = (Bundle) o2;
			return hasPrerequisite(base, other) ? 1 : -1;
		}
	};

	// the ordered list of plug-in descriptors for the plug-ins which
	// extend our extension point. this list should be in the
	// same order as plug-in activation
	private Set orderedPlugins = new TreeSet(comparator);

	// Mapping of plug-in descriptor to extension.
	private Map table = new HashMap();

	/**
	 * Helper method that creates a IStatus object given a message and a Throwable
	 * object.
	 * 
	 * @param message a human-readable message 
	 * @param exception a low-level exception, or <code>null</code> if not
	 *    applicable 
	 * @return the created status object
	 */
	private static IStatus createStatus(String message, Throwable exception) {
		return new Status(IStatus.ERROR, RestorerPlugin.PI_RESTORER, IStatus.ERROR, message, exception);
	}

	/**
	 * Call each of the restorers. Collect the results from each and return it.
	 */
	private IStatus callRestorers() {
		MultiStatus result = new MultiStatus(RestorerPlugin.PI_RESTORER, IStatus.OK, "Problems encountered while rebuilding workspace.", null); //$NON-NLS-1$

		// we will use the same timestamp for the rename of the plug-in
		// state locations to easily identify which ones were renamed at the same time.
		String timestamp = String.valueOf(System.currentTimeMillis());

		// call each of the restorers in order. the org.eclipse.core.resources restorer should
		// be the first one in the list so it can reconstruct the workspace tree
		for (Iterator i = orderedPlugins.iterator(); i.hasNext();) {
			Bundle bundle = (Bundle) i.next();
			IPath pluginStateLocation = Platform.getStateLocation(bundle);
			IPath corruptLocation = pluginStateLocation.removeTrailingSeparator().addFileExtension(timestamp);
			IExtension extension = (IExtension) table.get(bundle);
			try {
				// rename the state location to include a timestamp. then call the restorer
				// passing in the location of the corrupt state.
				if (pluginStateLocation.toFile().renameTo(corruptLocation.toFile())) {
					IRestorer restorer = (IRestorer) extension.getConfigurationElements()[0].createExecutableExtension(RestorerPlugin.RESTORER_CLASS_PROPERTY);
					result.add(restorer.restoreFrom(corruptLocation));
				} else
					result.add(createStatus("Unable to rename plug-in state location: " + pluginStateLocation, null)); //$NON-NLS-1$
			} catch (Throwable e) {
				// don't let anybody throw an exception.
				result.add(createStatus("Exception calling restorer code. (" + extension.getNamespace() + ")", e)); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
		return result;
	}

	/**
	 * Retrieves a list of contributors' extensions for the "restorers" 
	 * extension point.
	 * 
	 * @return the extensions
	 */
	private IExtension[] getExtensions() {
		IExtensionRegistry registry = Platform.getExtensionRegistry();
		IExtensionPoint extPt = registry.getExtensionPoint(RestorerPlugin.PT_RESTORERS);
		return extPt.getExtensions();
	}

	/**
	 * Returns <code>true</code> if the plug-in indicated by <code>base</code>
	 * contains the plug-in indicated by <code>other</code> in its prerequisite chain.
	 */
	protected boolean hasPrerequisite(Bundle base, Bundle other) {
		String requires = (String) base.getHeaders().get(Constants.REQUIRE_BUNDLE);
		try {
			ManifestElement[] prereqs = ManifestElement.parseHeader(Constants.REQUIRE_BUNDLE, requires);
			for (int i = 0; i < prereqs.length; i++) {
				String id = prereqs[i].getValue();
				if (id.equals(other.getSymbolicName()))
					return true;
				Bundle prereqBundle = Platform.getBundle(id);
				if (prereqBundle != null && hasPrerequisite(prereqBundle, other))
					return true;
			}
		} catch (BundleException e) {
			log(new Status(IStatus.ERROR, RestorerPlugin.PI_RESTORER, 1, "Error resolving prerequisite", e)); //$NON-NLS-1$
		}
		return false;
	}

	/**
	 * Calculate the order that we want to call the restorers.
	 * Should be based on the plug-in activation order.
	 */
	private void initialize() {
		IExtension[] extensions = getExtensions();
		for (int i = 0; i < extensions.length; i++) {
			IExtension extension = extensions[i];
			Bundle bundle = Platform.getBundle(extension.getNamespace());
			orderedPlugins.add(bundle);
			// add the association to the table
			table.put(bundle, extension);
		}
	}

	/** 
	 * Logs the provided status object in the plug-in log.
	 * 
	 * @param status the status to log
	 */
	private void log(IStatus status) {
		RestorerPlugin.getDefault().getLog().log(status);
	}

	/**
	 * @see IPlatformRunnable#run
	 */
	public Object run(Object args) throws Exception {
		initialize();
		IStatus result = callRestorers();
		if (result.isOK())
			return EXIT_OK;
		log(result);
		return new Integer(-1);
	}
}
